/* Plays KK Bossa on a piezo */
// Moved all arrays into program memory
// Added animation to a LCD

// Program Memory Library allows us store variables into program (flash) memory
// as opposed to static random access memory (SRAM or dynamic memory)
// refer to:http://www.nongnu.org/avr-libc/user-manual/group__avr__pgmspace.html
#include <avr/pgmspace.h>

// Tone Library allows us to play up to 3 tones on one piezo
// Note: In this code we only play 2 tones because a third tone will mess with timer 0;
//       and thus functions such as millis() and delay()
// Refer to its examples
#include <Tone.h> // includes pitches

// include the KK Bossa music
#include "KK_Bossa_Music.h"

// Liquid Crystal Library allow us to use an LCD
// Refer to its examples
#include <LiquidCrystal.h>

///* set up *///
const char switchPin = 12;
boolean start = 0; // default off

Tone Treble, Bass;
const char piezoPin1 = 10; // Treble
const char piezoPin2 = 11; // Bass

const char V0Pin = 5; // for fading the LCD text

const boolean GENDER = 0; // display male villager iff 1, display female villager iff 0

unsigned int skip = 0; // the measure to start playing from, note: starts from 1

// recall: We will be counting by 16th notes
// Based upon the bpm, determine how long one minibeat (16th note) lasts
const unsigned int ms_per_mb = 60000 / (KK_Bossa_bpm * 4); // milliseconds per minibeat

unsigned int num_mb; // song length in minibeats

// initialize the LCD
// lcd(RS,EN,D4,D5,D6,D7)
LiquidCrystal lcd(2, 3, 6, 7, 8, 9);

void setup() {
  Serial.begin(9600);

  pinMode(switchPin, INPUT);
  pinMode(V0Pin, OUTPUT);

  Treble.begin(piezoPin1);
  Bass.begin(piezoPin2);

  lcd.begin(16, 2); // 16x2 Liquid Crystal Display

  /// pre-start checks ///
  boolean passed_checks = true; // default true

  // check if array pairs have the same number of notes
  int Treble_Length = sizeof(KK_Bossa_Treble_Freq) / sizeof(KK_Bossa_Treble_Freq[0]);
  if (Treble_Length != sizeof(KK_Bossa_Treble_Notes) / sizeof(KK_Bossa_Treble_Notes[0])) {
    Serial.println(F("Error: Treble sizes are not the same"));
    passed_checks = false;
  }
  Serial.print(F("Number of notes in Treble: "));
  Serial.println(Treble_Length);

  int Bass_Length = sizeof(KK_Bossa_Bass_Freq) / sizeof(KK_Bossa_Bass_Freq[0]);
  if (Bass_Length != sizeof(KK_Bossa_Bass_Notes) / sizeof(KK_Bossa_Bass_Notes[0])) {
    Serial.println(F("Error: Bass sizes are not the same"));
    passed_checks = false;
  }
  Serial.print(F("Number of notes in Bass: "));
  Serial.println(Bass_Length);

  // check if the treble and base have the same number of minibeats
  int Treble_minibeats = 0;
  for (int i = 0; i < Treble_Length; i++) {
    Treble_minibeats = Treble_minibeats + (uint8_t)pgm_read_word(KK_Bossa_Treble_Notes + i); // necessary cast, not sure why. int, short, and long didn't work
  }
  int Bass_minibeats = 0;
  for (int i = 0; i < Bass_Length; i++) {
    Bass_minibeats = Bass_minibeats + (uint8_t)pgm_read_word(KK_Bossa_Bass_Notes + i);
  }
  if (Treble_minibeats != Bass_minibeats) {
    Serial.println(F("Error: Treble and Bass do not have the same number of beats"));
    passed_checks = false;
  }
  Serial.print(F("Number of minibeats in Treble: "));
  Serial.println(Treble_minibeats);
  Serial.print(F("Number of minibeats in Bass: "));
  Serial.println(Bass_minibeats);

  // halt code if failed to pass
  if (!passed_checks) {
    while (true);
  } else { // if passed, create global variable song length
    num_mb = Treble_minibeats; // total number of minibeats for entire song
    Serial.print(F("Song Length: ~"));
    int song_length_sec = num_mb * (ms_per_mb / 1000.0); // song length in seconds
    Serial.print(song_length_sec / 60);
    Serial.print(F(":"));
    Serial.print(song_length_sec % 60);
    Serial.println(F(" minutes"));
  }
}

///* main loop *///
void loop() {
  // wait until the button is pressed to play
  if (skip > 0) {
    Serial.print(F("Skip to Measure "));
    Serial.println(skip);
  }
  Serial.println(F("Ready to Play"));
  lcd.clear();
  lcd.setCursor(4, 0);
  lcd.print(F("Press to"));
  lcd.setCursor(5, 1);
  if (skip > 0) {
    lcd.print(F("Skip!"));
  } else {
    lcd.print(F("Start!"));
  }

  while (!start) {
    start = digitalRead(switchPin);
  }
  Serial.println(F("Playing..."));

  // need to keep track of the number of notes for Treble and Bass
  int n1 = 0; // index for Treble, default 0
  int n2 = 0; // index for Bass, default 0

  // will use countdowns to determine if it is time to play another note
  int c1 = ms_per_mb; // note countdown for Treble, default ms_per_mb
  int c2 = ms_per_mb; // note countdown for Bass, default ms_per_mb

  // play notes by minibeats
  // recall: each measure is 16 minibeats
  for (unsigned int mb = 0; mb < num_mb; mb++) {
    if ((skip > 0) && (mb %16 ==0)) {
      skip--; // decrease skip by one every measure (16 minibeats)
    }
    if (mb % 16 == 0) {
      Serial.println();
      Serial.print(F("Measure "));
      Serial.println(mb / 16);
    }
    Serial.print(mb);
    Serial.print(F(", "));

    // determine how long until the next note to change
    unsigned int t = MIN(ms_per_mb, c1, c2);

    if (skip == 0) {
      // do animation first before playing the next note
      // we want to make sure the notes play on time, so we animate inbetween minibeats
      unsigned int t0 = millis();
      Animation(mb, num_mb);
      unsigned int tf = millis() - t0; // the time it took to display the animation frame

      if (t - tf > 0) {
        delay(t - tf); // subtract the amount of time to display animation frame
      } else { // t-tf<0
        // animation took to long
        Serial.println();
        Serial.print("\tAnimation frame at minibeat ");
        Serial.print(mb);
        Serial.println(" took too long");
      }
    }

    if (t == c1) { // play the next treble note
      Treble.stop();
      c1 = (uint8_t)pgm_read_word(KK_Bossa_Treble_Notes + n1) * ms_per_mb;
      int freq1 = (uint16_t)pgm_read_word(KK_Bossa_Treble_Freq + n1);
      if (freq1 == 0) { // if it is a rest
        // do nothing
      } else { // play note
        if (skip == 0) {
          Treble.play(freq1);
        }
      }
      n1++;
    } else { // decrease c1
      c1 = c1 - ms_per_mb;
    }
    if (t == c2) { // play the next bass note
      Bass.stop();
      c2 = (uint8_t)pgm_read_word(KK_Bossa_Bass_Notes + n2) * ms_per_mb;
      int freq2 = (uint16_t)pgm_read_word(KK_Bossa_Bass_Freq + n2);
      if (freq2 == 0) { // if it is a rest
        // do nothing
      } else { // play note
        if (skip == 0) {
          Bass.play(freq2);
        }
      }
      n2++;
    } else { // decrease c2
      c2 = c2 - ms_per_mb;
    }
  }

  // turn off tones
  Treble.stop();
  Bass.stop();
  
  // turn contrast back on
  analogWrite(V0Pin, 0);

  // reset start
  start = 0;
  Serial.println(F("Finished"));
}

// subfunction MIN()
// returns the minimum of 3 integers
int MIN(int a, int b, int c) {
  int m = a;
  if (b < m) {
    m = b;
  }
  if (c < m) {
    return c;
  }
  return m;
}
